"""Module that holds the mapping of key press to action.
Holding that mapping, along with some metadata about each mapping (user
readable title and description of the action) in a data structure makes the
main keypress loop and generating a help/man document much simpler."""
import curses
from curses import ascii as curses_ascii
import types
import logging

from . import modes


CURSES_KEY_TO_STR = {key_int: key_str[4:] for key_str, key_int in
                     list(vars(curses).items())
                     if key_str.startswith('KEY_')}
# Add mapping for escape
CURSES_KEY_TO_STR[27] = 'ESC'


# Yeah, I know it has a lot of public methods, this is intended
# pylint: disable=too-many-public-methods
class ActionTable(object):
    """a class for accepting keys and performing actions"""
    def __init__(self, app_stdscr, scrobj):
        self.screen = scrobj
        self.app_stdscr = app_stdscr
        self._init_action_tables()

    # Not using self here, but I don't want this to be a staticmethod
    # pylint: disable=no-self-use
    def quit_action(self):
        """Quit the application"""
        # For now just exit
        exit(0)

    def resize_action(self):
        """Resize the screens on a resize event"""
        logging.info('RESIZE EVENT')
        self.screen.resize_windows(self.app_stdscr)
        self.screen.redraw_screen()

    def search_action(self):
        """Search downward through the current blame or commit"""
        if self._if_mode_normal() or self._if_mode_show():
            self.screen.get_search_string()
        else:
            self.action_mode_warning()

    def next_search_action(self):
        """Jump to the next search match (in the downward direction)"""
        if self._if_mode_normal() or self._if_mode_show():
            self.screen.jump_to_next_search_string()
        else:
            self.action_mode_warning()

    def prev_search_action(self):
        """Jump to the prev search match (in the upward direction)"""
        if self._if_mode_normal() or self._if_mode_show:
            self.screen.jump_to_previous_search_string()
        else:
            self.action_mode_warning()

    def vis_select_action(self):
        """Enter visual select mode (only from normal mode)"""
        if self._if_mode_normal():
            self.screen.init_vis_cursor()

    def show_line_commit_action(self):
        """Show a commit selected by the visual mode cursor"""
        if self._if_mode_visual():
            self.screen.init_git_show()

    def show_file_commit_action(self):
        """Show the current revision commit"""
        self.screen.init_git_show_file()

    def parent_action(self):
        """Open a new git blame to the parent of the current commit"""
        self.screen.add_blame_parent()

    def ancestor_action(self):
        """Open a new git blame to the ancestor of the current commit"""
        self.screen.add_blame_ancestor()

    def drill_action(self):
        """Drill down past the commit highlighted in visual mode.
           Opens a new git blame"""
        if self._if_mode_visual():
            self.screen.add_blame_drill()

    def pop_action(self):
        """Pop back to previous git object"""
        self.screen.restore_prev_content_and_state()

    def move_up_action(self):
        """move the screen or visual select cursor up"""
        # if mode is visual, move line highlight
        if self._if_mode_visual():
            self.screen.move_cursor_up()
        # if mode is normal, move screen
        else:
            self.screen.move_scr_up()
        return True

    def move_down_action(self):
        """move the screen or visual select cursor down"""
        # if mode is visual, move line highlight
        if self._if_mode_visual():
            self.screen.move_cursor_down()
        # if mode is normal, move screen
        else:
            self.screen.move_scr_down()
        return True

    def page_up_action(self):
        """Move the screen up one page"""
        self.screen.move_scr_up_page()

    def page_down_action(self):
        """Move the screen down one page"""
        self.screen.move_scr_down_page()

    def page_right_action(self):
        """Move the screen half a page to the right"""
        self.screen.move_scr_right_hpage()

    def page_left_action(self):
        """Move the screen half a page to the left"""
        self.screen.move_scr_left_hpage()

    def jump_top_action(self):
        """Jump to the top of the screen"""
        self.screen.move_scr_to_top()

    def jump_bottom_action(self):
        """Jump to the bottom of the screen"""
        self.screen.move_scr_to_bottom()

    def help_action(self):
        """Display the help message"""
        if self._if_mode_help():
            self.screen.set_status_bar_next_msg_bold('Already displaying '
                                                     'help!')
        else:
            help_msg = self.generate_help()
            self.screen.display_help(help_msg)

    def escape_action(self):
        """Return to Normal mode"""
        if self.screen.mode in [modes.MODE_VISUAL]:
            self.screen.mode = modes.MODE_NORMAL
        elif self.screen.mode in [modes.MODE_SHOW, modes.MODE_HELP]:
            self.screen.restore_prev_content_and_state()

    def _if_mode_normal(self):
        """check if the screen obj is in NORMAL mode"""
        return self.screen.mode == modes.MODE_NORMAL

    def _if_mode_visual(self):
        """check if the screen obj is in VISUAL mode"""
        return self.screen.mode == modes.MODE_VISUAL

    def _if_mode_show(self):
        """check if the screen obj is in SHOW mode"""
        return self.screen.mode == modes.MODE_SHOW

    def _if_mode_help(self):
        """check if the screen obj is in HELP mode"""
        return self.screen.mode == modes.MODE_HELP

    def action_mode_warning(self):
        """Warn the user that the action they are trying to perform isn't
        available in the current mode they're in"""
        self.screen.set_status_bar_next_msg_bold('Action not available in '
                                                 '%s mode!'
                                                 % self.screen.mode)

    def process_key(self, key):
        """Take a key and check if we have a corresponding action, if so call
        the action function, and log the keypress"""
        isascii = curses_ascii.isascii(key)
        ascii_key = chr(key) if isascii else None
        logging.info('Actions - process_key: KEY PRESSED (int, ascii), '
                     '(%s, %s)', key, ascii_key)
        action = self.key_to_action.get(key, None)
        if action:
            logging.info('Actions. process_key: Action func: %s',
                         str(action['action_name']))
            return action['action_func']()
        else:
            logging.info('Actions - process_key: Key not in action table')

    def generate_help(self):
        """Automagically generate a help document from the action descriptions
        in the action table"""

        def stringify_key(key):
            """Convert key ints into curses strings for displaying in the help
            text."""
            if isinstance(key, int):
                key_str = CURSES_KEY_TO_STR.get(key)
                if key_str:
                    return key_str
                else:
                    return ''
            return key
        help_lines = []
        help_lines.append('KEYS: ACTION - DESCRIPTION')
        help_lines.append('--------------------------')
        for keys, action in self.keys_to_action:
            if action['show_in_help_msg']:
                keys_strs = [stringify_key(key) for key in keys]
                keys_str = ', '.join([_f for _f in keys_strs if _f])
                action_name = action['action_name']
                action_desc = action['action_desc']
                help_lines.append(' %s:   %s' % (keys_str, action_name))
                help_lines.append('    %s' % (action_desc))
                help_lines.append('')
        return help_lines

    @staticmethod
    def _create_action_entry(action_name, action_func, action_desc='',
                             show_in_help_msg=True):
        """Given inputs, validate they are at least of the right type and
        create a action table entry
         Schema of a single entry:
         <keys_list>: {'action_name': <name of action str>,
                       'acton_desc': <short description of action str>,
                       'action_func': <action_function to call>
                       'show_in_help_msg': <bool>
                       }
        """
        # If no description was passed use the doc of the function
        if not action_desc:
            action_desc = action_func.__doc__
        assert isinstance(action_name, str)
        assert isinstance(action_desc, str)
        assert isinstance(action_func, types.MethodType)
        assert isinstance(show_in_help_msg, bool)

        # If no action provided use func doc
        if not action_desc:
            action_desc = action_func.__doc__

        # sanitize the desc to have no new lines or spaces/margins
        clean_lines = [x.strip() for x in action_desc.splitlines()]

        return {'action_name': action_name,
                'action_desc': ' '.join(clean_lines),
                'action_func': action_func,
                'show_in_help_msg': show_in_help_msg}

    KEYS_TUPLE_TO_ACTION2 = []

    def _init_action_tables(self):
        """Generate a list of tuples that map keys to an action that they
        perform"""
        self.keys_to_action = [
            (['q'],
             self._create_action_entry('Quit', self.quit_action)),
            ([curses.KEY_RESIZE],
             self._create_action_entry('Screen Resize', self.resize_action,
                                       show_in_help_msg=False)),
            (['/'],
             self._create_action_entry('Search', self.search_action)),
            (['n'],
             self._create_action_entry('Next Search Match',
                                       self.next_search_action)),
            (['N'],
             self._create_action_entry('Prev Search Match',
                                       self.prev_search_action)),
            (['v', 's'],
             self._create_action_entry('Visual Select Mode',
                                       self.vis_select_action)),
            (['o'],
             self._create_action_entry('Show/View Commit',
                                       self.show_line_commit_action)),

            (['O'],
             self._create_action_entry('Show/View file Commit',
                                       self.show_file_commit_action)),

            ([27],
             self._create_action_entry('Normal Mode',
                                       self.escape_action)),
            ([10, curses.KEY_ENTER, 'd'],
             self._create_action_entry('Drill Down', self.drill_action)),
            (['<', ','],
             self._create_action_entry('Parent blame', self.parent_action)),
            (['>', '.'],
             self._create_action_entry('Ancestor blame',
                                       self.ancestor_action)),
            ([8, curses.KEY_BACKSPACE, curses.KEY_DC, 127, 'f'],
             self._create_action_entry('Pop Back', self.pop_action)),
            (['k', curses.KEY_UP],
             self._create_action_entry('Move Up',
                                       self.move_up_action,
                                       show_in_help_msg=False)),
            (['j', curses.KEY_DOWN],
             self._create_action_entry('Move Down',
                                       self.move_down_action,
                                       show_in_help_msg=False)),
            ([curses.KEY_PPAGE],
             self._create_action_entry('Move Up Page',
                                       self.page_up_action,
                                       show_in_help_msg=False)),
            ([curses.KEY_NPAGE],
             self._create_action_entry('Move Down Page',
                                       self.page_down_action,
                                       show_in_help_msg=False)),
            ([curses.KEY_RIGHT],
             self._create_action_entry('Move Right Page',
                                       self.page_right_action,
                                       show_in_help_msg=False)),
            ([curses.KEY_LEFT],
             self._create_action_entry('Move Left Page',
                                       self.page_left_action,
                                       show_in_help_msg=False)),
            (['g', curses.KEY_HOME],
             self._create_action_entry('Jump to Top',
                                       self.jump_top_action,
                                       show_in_help_msg=False)),
            (['G', curses.KEY_END],
             self._create_action_entry('Jump to Bottom',
                                       self.jump_bottom_action)),
            (['h'], self._create_action_entry('Help', self.help_action)),
        ]

        # Create a second mapping of each single key (as its int keycode) to
        # the action, to make single key lookups easier in main loop
        self.key_to_action = {}
        for keys, action in self.keys_to_action:
            for key in keys:
                if isinstance(key, int):
                    self.key_to_action[key] = action
                elif isinstance(key, str):
                    # convert char to int key code
                    assert len(key) == 1, "Keys must be a single char!"
                    self.key_to_action[ord(key)] = action



"""Classes and utility functions related to git data and operations on it"""
import os
import gc
import itertools
import curses
import zlib

import sh

from .util import buildlinesdict


class BadRevException(Exception):
    """Simple custom exception to make exception handling easier up the
    stack"""
    pass


class NoSuchRevException(Exception):
    """Simple custom exception to make exception handling easier up the
    stack"""
    pass

ABBREV_LEN = 8
# Git blame appears to show always n+1 chararcters of the shaw, where n is
# the number of chars you actually asked for...
ABBREV_LEN_BLAME = ABBREV_LEN - 1
# Git color pair names
GIT_COLOR_OLD = 8
GIT_COLOR_NEW = 9
GIT_COLOR_FRAG = 10
GIT_COLOR_FUNC = 11
GIT_COLOR_SHOW = 12
GIT_COLOR_META = 13
# Git hunk constants
GIT_HUNK_MODIFIED_ADDED = 'ma'
GIT_HUNK_MODIFIED_REMOVED = 'mr'


def init_git_color_pairs():
    """ init git color pairs
    Note: -1 maps to default background color"""
    curses.init_pair(GIT_COLOR_OLD, curses.COLOR_RED, -1)
    curses.init_pair(GIT_COLOR_NEW, curses.COLOR_GREEN, -1)
    curses.init_pair(GIT_COLOR_FRAG, curses.COLOR_CYAN, -1)
    curses.init_pair(GIT_COLOR_FUNC, curses.COLOR_BLUE, -1)
    curses.init_pair(GIT_COLOR_SHOW, curses.COLOR_YELLOW, -1)
    curses.init_pair(GIT_COLOR_META, curses.COLOR_MAGENTA, -1)


def get_head_rev(filename):
    """Get the revision pointed to by HEAD"""
    cmd = ['rev-list', '--abbrev=%d' % ABBREV_LEN, '--max-count=1',
           'HEAD', '--', filename]
    return get_commit_helper(cmd)


def get_commit_helper(cmds):
    """Get the ancestor of <current_rev>"""
    git = sh.git.bake(_cwd=os.getcwd())
    revlist = git(*cmds)
    if not revlist.splitlines():
        raise NoSuchRevException()
    else:
        # return the firs 8 chars of the revision
        return revlist.splitlines()[-1][:8]


# pylint: disable=too-few-public-methods
class GitLog(object):
    """A class to represent the git log"""

    class LogEntry(object):
        """A class object to represent an entry of the git log.
        Stores the sha and filename for that commit, and references to this
        commits parent and ancestor"""
        # pylint: disable=too-many-arguments
        def __init__(self, sha, filename, desc, parent=None, ancestor=None):
            self.sha = sha
            self.filename = filename
            self.desc = desc
            self.parent = parent
            self.ancestor = ancestor

    def __init__(self, initial_filename):
        self.gitlog = {}
        git = sh.git.bake('--no-pager', _cwd=os.getcwd())
        initial_log = git.log('--no-color', '--abbrev=%d' % ABBREV_LEN,
                              '--follow', '--oneline', '--name-only', '--',
                              '%s' % initial_filename)
        previous_log_entry = None
        for sha_and_desc, filename in zip(*[iter(initial_log.splitlines())]*2):
            sha = sha_and_desc.split()[0]
            desc = ' '.join(sha_and_desc.split()[1:])
            curr_log_entry = self.LogEntry(sha, filename, desc,
                                           ancestor=previous_log_entry)
            if previous_log_entry:
                previous_log_entry.parent = curr_log_entry
            self.gitlog[sha] = curr_log_entry
            previous_log_entry = curr_log_entry

    def __getitem__(self, item):
        return self.gitlog[item]

    def __str__(self):
        return self.gitlog.__str__()

    def __iter__(self):
        return self.gitlog.__iter__()

    def __len__(self):
        return self.gitlog.__len__()


class Blame(object):
    """A class to represent a blame.

    Attributes:
        git_sha - The SHA of the git commit being blamed if one
        filename - The name of the file we are blaming
        lines - The output of git blame by line in a default dict
    """
    def __init__(self, filename, git_sha=''):
        git = sh.git.bake('--no-pager', _cwd=os.getcwd())
        self.filename = filename
        if git_sha:
            try:
                cmd = [git_sha, '--abbrev=%d' % ABBREV_LEN_BLAME,
                       '--', filename]
                self.lines = buildlinesdict(
                    git.blame(*cmd).stdout.splitlines())
            except sh.ErrorReturnCode_128 as exc:
                if "no such path" in exc.message:
                    raise NoSuchRevException(exc.stderr)
                if "bad revision" in exc.message:
                    raise BadRevException(exc.stderr)
                raise
            self.git_sha = git_sha
        else:
            self.git_sha = get_head_rev(filename)
            self.lines = buildlinesdict(git.blame(
                '--abbrev=%d' % ABBREV_LEN_BLAME,
                filename).stdout.splitlines())
        # lines will grow if any '~' padding is added
        self.numlines = len(self.lines)
        del git
        gc.collect()

    def compress(self):
        """Compress the lines attribute which stores the content of the git
        blame. This makes storing each stack frame cheaper"""
        blame_text = '\n'.join([line[0] for line in self.lines.values()])
        compressed_lines = zlib.compress(blame_text.encode('utf-8'))
        del self.lines
        gc.collect()
        self.lines = compressed_lines

    def decompress(self):
        """Decompress the lines attribute which stores the content of the git
        blame."""
        temp_lines = zlib.decompress(self.lines)
        self.lines = buildlinesdict(temp_lines.splitlines())
        del temp_lines
        gc.collect()

    def __len__(self):
        return self.numlines

    def __getitem__(self, index):
        return self.lines[index]

    def __setitem__(self, index, value):
        self.lines[index] = value


class Show(object):
    """A class to represent a git show.

    Attributes:
        showobj - The function object returned by sh.git, contains attributes
                   like the return code of the command, the stdout, etc.
        git_sha - The SHA of the git commit being showed
    """
    def __init__(self, git_sha):
        git = sh.git.bake('--no-pager', _cwd=os.getcwd())
        showobj = git.show('--no-color', git_sha)
        self.git_sha = git_sha
        self.lines = buildlinesdict(showobj.stdout.splitlines())
        self.colorize_lines()
        del git
        del showobj
        gc.collect()

    def colorize_lines(self):
        """Add the necessary curses attributes to each line to produce the
        desired color, highlight, etc"""
        iter_lines = self.lines.items()
        for line_idx, (line_str, line_attrs) in iter_lines:
            # commit header
            if line_str.startswith('commit'):
                self.lines[line_idx] = (line_str,
                                        curses.color_pair(GIT_COLOR_SHOW))

            # meta information block
            elif line_str.startswith('diff'):
                self.lines[line_idx] = (line_str,
                                        curses.color_pair(GIT_COLOR_META))
                for line_idx, (line_str,
                               line_attrs) in itertools.islice(iter_lines, 3):
                    self.lines[line_idx] = (line_str,
                                            curses.A_DIM |
                                            curses.color_pair(GIT_COLOR_META))
            # hunk/fragment header
            elif line_str.startswith('@@'):
                # TODO: tease out function in hunk header
                self.lines[line_idx] = (line_str,
                                        curses.color_pair(GIT_COLOR_FRAG))

            # added lines
            elif line_str.startswith('+'):
                self.lines[line_idx] = (line_str,
                                        curses.color_pair(GIT_COLOR_NEW))

            # removed lines
            elif line_str.startswith('-'):
                self.lines[line_idx] = (line_str,
                                        curses.color_pair(GIT_COLOR_OLD))

    def __len__(self):
        return len(self.lines)

    def __getitem__(self, index):
        return self.lines[index]

    def __setitem__(self, index, value):
        self.lines[index] = value



"""Module with the implementation and related functions for the content
stack"""
from collections import namedtuple

from . import git


class ContentStack(object):
    """A class to represent the content added as users drill into git revs,
    cycle through git history, show commits, etc"""
    def __init__(self):
        self.content_stack = []

    # These frames also store the state necessary to move back to the
    # previous frame/screen when it is popped. Essentially storing a snapshot
    # of what the screen looked like before we add this frame
    StackFrame = namedtuple('StackFrame', ['content', 'last_current_line',
                                           'last_current_width',
                                           'last_cursor_line', 'last_mode'])

    def __len__(self):
        return len(self.content_stack)

    def peek(self):
        """Peek and return the content on the top of the stack,
        without removing it from the stack"""
        return self.content_stack[-1]

    def pop(self):
        """Pop content off the content stack, first checking if we have any
        extra content that can be popped (I.e. we don't want to be left]
        without any content to display to the user."""
        cstack = self.content_stack
        if len(cstack) <= 1:
            raise IndexError('No content to pop')
        frame_to_return = cstack.pop()
        content_to_restore = self.content_stack[-1].content
        # if content is a blame then decompress it (See comment below for
        # compression)
        if isinstance(content_to_restore, git.Blame):
            self.content_stack[-1].content.decompress()
        return frame_to_return

    # pylint: disable=too-many-arguments
    def add(self, newcontent_obj, current_line, current_width,
            cursor_line, mode):
        """Add a new frame object to the Content Stack, preserving some of our
        current state as we do that, so it can be restored when we pop this."""
        # compress current stack before adding the new stack
        if len(self.content_stack) >= 1:
            last_content = self.content_stack[-1].content
            # If the last content was a blame, then compress it as these can
            # be quite large.
            # TODO: Make a base class in git.py that will allow both Blame and
            # Show objects to inherit the compressing functions, because we
            # might as well compress the git show as well, they can sometimes
            # be large.
            if isinstance(last_content, git.Blame):
                last_content.compress()
        stack_frame = self.StackFrame(newcontent_obj, current_line,
                                      current_width, cursor_line, mode)
        self.content_stack.append(stack_frame)



"""Different modes of the bblame application"""
# Normal/standard view
MODE_NORMAL = 'normal'
# Visual select mode for selecting lines of the blame file
MODE_VISUAL = 'visual'
# Viewing a show not a blame
MODE_SHOW = 'show'
# Viewing a help message
MODE_HELP = 'help'



"""Utility functions and objects"""
from collections import defaultdict
import curses


def buildlinesdict(lines, defaultstr='~', default_attr=curses.A_NORMAL):
    """build a default dict with schema:
    index --> (line or <defaultstr>, curses str attributes)
    The default string makes scrolling past the edge of the file much more
    graceful.
    I.e. having a default value rather than handling IndexError and then
    returning the default string.
    This function will add a default or the provided attribute to each line.
    The lines can be further decorated at later times by logically OR-ing in
    other attributes to the attribute set here."""
    ret = defaultdict(lambda: (defaultstr, default_attr))
    for idx, bline in enumerate(lines):
        if isinstance(bline, bytes):
            bline = bline.decode('UTF-8')
        ret[idx] = (bline, default_attr)
    return ret


# In this case I really do want an object pylint, but thanks...
# pylint: disable=R0903
class BidirectionalCycle(object):
    """A cycle iterator that can iterate in both directions (e.g. has next
    and prev).
    This is a simple object that supports the iterator protocol but it doesn't
    behave like one might expect a standard iterator to (e.g. a generator that
    lazily produces the next value) this object will keep the WHOLE LIST in
    memory, so use WITH CAUTION"""
    def __init__(self, list_seq, starting_index=0, no_wrap=False):
        self.current_index = self.init_index = starting_index
        # CURRENTLY ONLY SUPPORT LISTS
        assert isinstance(list_seq, list), 'Currently only supports lists'
        self.seq = list_seq
        self.no_wrap = no_wrap
        self.start_of_day = True

    def next(self):
        """Maintain support for python2 iterator"""
        return self.__next__()

    def __next__(self):
        """return the next item in the iteration"""
        self._check_len()
        if self.start_of_day:
            return self._start_of_day()

        self._move_index_next()
        next_item = self.seq[self.current_index]

        return next_item

    def prev(self):
        """return the previous item in the iteration"""
        self._check_len()
        if self.start_of_day:
            if self.no_wrap:
                return ''
            return self._start_of_day()

        self._move_index_prev()
        prev_item = self.seq[self.current_index]

        return prev_item

    def curr(self):
        """Returns the current item in the iteration"""
        self._check_len()
        if self.start_of_day:
            return self._start_of_day()

        return self.seq[self.current_index]

    def _move_index_next(self):
        """Move the index in the next direction"""
        # check if we need to wrap around to the beginning
        if self.current_index == len(self.seq) - 1:
            if self.no_wrap:
                return
            self.current_index = 0
        else:
            self.current_index = self.current_index + 1

    def _move_index_prev(self):
        """Move the index in the prev direction"""
        # check if we need to wrap around to the end
        if self.current_index == 0:
            if self.no_wrap:
                return
            self.current_index = len(self.seq) - 1
        else:
            self.current_index = self.current_index - 1

    def _check_len(self):
        """As itertools.cycle does, raise StopIteration if the sequence
        is empty"""
        if len(self.seq) == 0:
            raise StopIteration

    def _start_of_day(self):
        """print out the init_index of the sequence and set start_of_day to
        false"""
        self.start_of_day = False
        return self.seq[self.init_index]

    def __str__(self):
        return str(self.seq)

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.seq)

    def __contains__(self, item):
        return item in self.seq
